#!/usr/bin/env python
# coding: latin-1
"""
This module is designed to communicate with the PicoBorg Reverse

Use by creating an instance of the class, call the Init function, then command as desired, e.g.
import PicoBorgRev
PBR = PicoBorgRev.PicoBorgRev()
PBR.Init()
# User code here, use PBR to control the board

Multiple boards can be used when configured with different IC addresses by creating multiple instances, e.g.
import PicoBorgRev
PBR1 = PicoBorgRev.PicoBorgRev()
PBR2 = PicoBorgRev.PicoBorgRev()
PBR1.i2cAddress = 0x44
PBR2.i2cAddress = 0x45
PBR1.Init()
PBR2.Init()
# User code here, use PBR1 and PBR2 to control each board separately

For explanations of the functions available call the Help function, e.g.
import PicoBorgRev
PBR = PicoBorgRev.PicoBorgRev()
PBR.Help()
See the website at www.piborg.org/picoborgreverse for more details
"""

# Import the libraries we need
import smbus
import types
import time

# Constant values
PWM_MAX                 = 255
I2C_MAX_LEN             = 4

I2C_ID_PICOBORG_REV     = 0x15

COMMAND_ALL_OFF         = 9     # Switch everything off
COMMAND_RESET_EPO       = 10    # Resets the EPO flag, use after EPO has been tripped and switch is now clear
COMMAND_GET_EPO         = 11    # Get the EPO latched flag
COMMAND_SET_EPO_IGNORE  = 12    # Set the EPO ignored flag, allows the system to run without an EPO
COMMAND_GET_EPO_IGNORE  = 13    # Get the EPO ignored flag
COMMAND_GET_DRIVE_FAULT = 14    # Get the drive fault flag, indicates faults such as short-circuits and under voltage
COMMAND_SET_ALL_FWD     = 15    # Set all motors PWM rate in a forwards direction
COMMAND_SET_ALL_REV     = 16    # Set all motors PWM rate in a reverse direction
COMMAND_SET_FAILSAFE    = 17    # Set the failsafe flag, turns the motors off if communication is interrupted
COMMAND_GET_FAILSAFE    = 18    # Get the failsafe flag
COMMAND_SET_ENC_MODE    = 19    # Set the board into encoder or speed mode
COMMAND_GET_ENC_MODE    = 20    # Get the boards current mode, encoder or speed

COMMAND_GET_EPO_CURRENT_STATE    = 32    # Get the Emergency button state

#PID variables
COMMAND_SET_KMOT      = 21    # kmot
COMMAND_SET_KP      = 22    # 
COMMAND_SET_KI      = 23    # 
COMMAND_GET_KMOT      = 24    # 
COMMAND_GET_KP    = 25    # 
COMMAND_GET_KI    = 26    # 

COMMAND_GET_ENC_MOVING  = 27    # Get the status of encoders moving
COMMAND_SET_ENC_SPEED   = 28    # Set the maximum PWM rate in encoder mode
COMMAND_GET_ENC_SPEED   = 29    # Get the maximum PWM rate in encoder mode
COMMAND_GET_ID          = 0x99  # Get the board identifier
COMMAND_SET_I2C_ADD     = 0xAA  # Set a new I2C address

COMMAND_SET_SPEED_FBCK  = 30
COMMAND_GET_SPEED_FBCK  = 31

COMMAND_VALUE_FWD       = 1     # I2C value representing forward
COMMAND_VALUE_REV       = 2     # I2C value representing reverse

COMMAND_VALUE_ON        = 1     # I2C value representing on
COMMAND_VALUE_OFF       = 0     # I2C value representing off


def ScanForPicoBorgReverse(busNumber = 1):
    """
ScanForPicoBorgReverse([busNumber])

Scans the IC bus for a PicoBorg Reverse boards and returns a list of all usable addresses
The busNumber if supplied is which IC bus to scan, 0 for Rev 1 boards, 1 for Rev 2 boards, if not supplied the default is 1
    """
    found = []
    print 'Scanning IC bus #%d' % (busNumber)
    bus = smbus.SMBus(busNumber)
    for address in range(0x03, 0x78, 1):
        try:
            i2cRecv = bus.read_i2c_block_data(address, COMMAND_GET_ID, I2C_MAX_LEN)
            if len(i2cRecv) == I2C_MAX_LEN:
                if i2cRecv[1] == I2C_ID_PICOBORG_REV:
                    print 'Found PicoBorg Reverse at %02X' % (address)
                    found.append(address)
                else:
                    pass
            else:
                pass
        except KeyboardInterrupt:
            raise
        except:
            pass
    if len(found) == 0:
        print 'No PicoBorg Reverse boards found, is bus #%d correct (should be 0 for Rev 1, 1 for Rev 2)' % (busNumber)
    elif len(found) == 1:
        print '1 PicoBorg Reverse board found'
    else:
        print '%d PicoBorg Reverse boards found' % (len(found))
    return found


def SetNewAddress(newAddress, oldAddress = -1, busNumber = 1):
    """
SetNewAddress(newAddress, [oldAddress], [busNumber])

Scans the IC bus for the first PicoBorg Reverse and sets it to a new I2C address
If oldAddress is supplied it will change the address of the board at that address rather than scanning the bus
The busNumber if supplied is which IC bus to scan, 0 for Rev 1 boards, 1 for Rev 2 boards, if not supplied the default is 1
Warning, this new IC address will still be used after resetting the power on the device
    """
    if newAddress < 0x03:
        print 'Error, IC addresses below 3 (0x03) are reserved, use an address between 3 (0x03) and 119 (0x77)'
        return
    elif newAddress > 0x77:
        print 'Error, IC addresses above 119 (0x77) are reserved, use an address between 3 (0x03) and 119 (0x77)'
        return
    if oldAddress < 0x0:
        found = ScanForPicoBorgReverse(busNumber)
        if len(found) < 1:
            print 'No PicoBorg Reverse boards found, cannot set a new IC address!'
            return
        else:
            oldAddress = found[0]
    print 'Changing IC address from %02X to %02X (bus #%d)' % (oldAddress, newAddress, busNumber)
    bus = smbus.SMBus(busNumber)
    try:
        i2cRecv = bus.read_i2c_block_data(oldAddress, COMMAND_GET_ID, I2C_MAX_LEN)
        if len(i2cRecv) == I2C_MAX_LEN:
            if i2cRecv[1] == I2C_ID_PICOBORG_REV:
                foundChip = True
                print 'Found PicoBorg Reverse at %02X' % (oldAddress)
            else:
                foundChip = False
                print 'Found a device at %02X, but it is not a PicoBorg Reverse (ID %02X instead of %02X)' % (oldAddress, i2cRecv[1], I2C_ID_PICOBORG_REV)
        else:
            foundChip = False
            print 'Missing PicoBorg Reverse at %02X' % (oldAddress)
    except KeyboardInterrupt:
        raise
    except:
        foundChip = False
        print 'Missing PicoBorg Reverse at %02X' % (oldAddress)
    if foundChip:
        bus.write_byte_data(oldAddress, COMMAND_SET_I2C_ADD, newAddress)
        time.sleep(0.1)
        print 'Address changed to %02X, attempting to talk with the new address' % (newAddress)
        try:
            i2cRecv = bus.read_i2c_block_data(newAddress, COMMAND_GET_ID, I2C_MAX_LEN)
            if len(i2cRecv) == I2C_MAX_LEN:
                if i2cRecv[1] == I2C_ID_PICOBORG_REV:
                    foundChip = True
                    print 'Found PicoBorg Reverse at %02X' % (newAddress)
                else:
                    foundChip = False
                    print 'Found a device at %02X, but it is not a PicoBorg Reverse (ID %02X instead of %02X)' % (newAddress, i2cRecv[1], I2C_ID_PICOBORG_REV)
            else:
                foundChip = False
                print 'Missing PicoBorg Reverse at %02X' % (newAddress)
        except KeyboardInterrupt:
            raise
        except:
            foundChip = False
            print 'Missing PicoBorg Reverse at %02X' % (newAddress)
    if foundChip:
        print 'New IC address of %02X set successfully' % (newAddress)
    else:
        print 'Failed to set new IC address...'


# Class used to control PicoBorg Reverse
class PicoBorgRev:
    """
This module is designed to communicate with the PicoBorg Reverse

busNumber               IC bus on which the PicoBorg Reverse is attached (Rev 1 is bus 0, Rev 2 is bus 1)
bus                     the smbus object used to talk to the IC bus
i2cAddress              The IC address of the PicoBorg Reverse chip to control
foundChip               True if the PicoBorg Reverse chip can be seen, False otherwise
printFunction           Function reference to call when printing text, if None "print" is used
    """

    # Shared values used by this class
    busNumber               = 1     # Check here for Rev 1 vs Rev 2 and select the correct bus
    i2cAddress              = 0x44  # IC address, override for a different address
    bus                     = None
    foundChip               = False
    printFunction           = None


    def Print(self, message):
        """
Print(message)

Wrapper used by the PicoBorgRev instance to print messages, will call printFunction if set, print otherwise
        """
        if self.printFunction == None:
            print message
        else:
            self.printFunction(message)


    def NoPrint(self, message):
        """
NoPrint(message)

Does nothing, intended for disabling diagnostic printout by using:
PBR = PicoBorgRev.PicoBorgRev()
PBR.printFunction = PBR.NoPrint
        """
        pass


    def Init(self, tryOtherBus = True):
        """
Init([tryOtherBus])

Prepare the I2C driver for talking to the PicoBorg Reverse
If tryOtherBus is True or omitted, this function will attempt to use the other bus if the PicoBorg Reverse devices can not be found on the current busNumber
        """
        self.Print('Loading PicoBorg Reverse on bus %d, address %02X' % (self.busNumber, self.i2cAddress))

        # Open the bus
        self.bus = smbus.SMBus(self.busNumber)

        # Check for PicoBorg Reverse
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_ID, I2C_MAX_LEN)
            if len(i2cRecv) == I2C_MAX_LEN:
                if i2cRecv[1] == I2C_ID_PICOBORG_REV:
                    self.foundChip = True
                    self.Print('Found PicoBorg Reverse at %02X' % (self.i2cAddress))
                else:
                    self.foundChip = False
                    self.Print('Found a device at %02X, but it is not a PicoBorg Reverse (ID %02X instead of %02X)' % (self.i2cAddress, i2cRecv[1], I2C_ID_PICOBORG_REV))
            else:
                self.foundChip = False
                self.Print('Missing PicoBorg Reverse at %02X' % (self.i2cAddress))
        except KeyboardInterrupt:
            raise
        except:
            self.foundChip = False
            self.Print('Missing PicoBorg Reverse at %02X' % (self.i2cAddress))

        # See if we are missing chips
        if not self.foundChip:
            self.Print('PicoBorg Reverse was not found')
            if tryOtherBus:
                if self.busNumber == 1:
                    self.busNumber = 0
                else:
                    self.busNumber = 1
                self.Print('Trying bus %d instead' % (self.busNumber))
                self.Init(False)
            else:
                self.Print('Are you sure your PicoBorg Reverse is properly attached, the correct address is used, and the I2C drivers are running?')
                self.bus = None
        else:
            self.Print('PicoBorg Reverse loaded on bus %d' % (self.busNumber))


    def ResetEpo(self):
        """
ResetEpo()

Resets the EPO latch state, use to allow movement again after the EPO has been tripped
        """
        try:
            self.bus.write_byte_data(self.i2cAddress, COMMAND_RESET_EPO, 0)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed resetting EPO!')


    def GetEpo(self):
        """
state = GetEpo()

Reads the system EPO latch state.
If False the EPO has not been tripped, and movement is allowed.
If True the EPO has been tripped, movement is disabled if the EPO is not ignored (see SetEpoIgnore)
    Movement can be re-enabled by calling ResetEpo.
        """ 
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_EPO, I2C_MAX_LEN)
            print "GetEpo::i2cRecv=",i2cRecv
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading EPO ignore state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True


    def SetEpoIgnore(self, state):
        """
SetEpoIgnore(state)

Sets the system to ignore or use the EPO latch, set to False if you have an EPO switch, True if you do not
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        try:
            self.bus.write_byte_data(self.i2cAddress, COMMAND_SET_EPO_IGNORE, level)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending EPO ignore state!')


    def GetEpoIgnore(self):
        """
state = GetEpoIgnore()

Reads the system EPO ignore state, False for using the EPO latch, True for ignoring the EPO latch
        """ 
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_EPO_IGNORE, I2C_MAX_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading EPO ignore state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def GetEpoCurrentState(self):
        """


Reads the system EPO current state.
.
        """ 
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_EPO_CURRENT_STATE, I2C_MAX_LEN)
            print "GetEpoCurrentState:: i2cRecv=",i2cRecv
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading EPO Current state ignore state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            
            return False
        else:
           
            return True



    def SetCommsFailsafe(self, state):
        """
SetCommsFailsafe(state)

Sets the system to enable or disable the communications failsafe
The failsafe will turn the motors off unless it is commanded at least once every 1/4 of a second
Set to True to enable this failsafe, set to False to disable this failsafe
The failsafe is disabled at power on
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        try:
            self.bus.write_byte_data(self.i2cAddress, COMMAND_SET_FAILSAFE, level)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending communications failsafe state!')


    def GetCommsFailsafe(self):
        """
state = GetCommsFailsafe()

Read the current system state of the communications failsafe, True for enabled, False for disabled
The failsafe will turn the motors off unless it is commanded at least once every 1/4 of a second
        """ 
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_FAILSAFE, I2C_MAX_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading communications failsafe state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True


    def GetDriveFault(self):
        """
state = GetDriveFault()

Reads the system drive fault state, False for no problems, True for a fault has been detected
Faults may indicate power problems, such as under-voltage (not enough power), and may be cleared by setting a lower drive power
If a fault is persistent, it repeatably occurs when trying to control the board, this may indicate a wiring problem such as:
    * The supply is not powerful enough for the motors
        The board has a bare minimum requirement of 6V to operate correctly
        A recommended minimum supply of 7.2V should be sufficient for smaller motors
    * The + and - connections for either motor are connected to each other
    * Either + or - is connected to ground (GND, also known as 0V or earth)
    * Either + or - is connected to the power supply (V+, directly to the battery or power pack)
    * One of the motors may be damaged
Faults will self-clear, they do not need to be reset, however some faults require both motors to be moving at less than 100% to clear
The easiest way to check is to put both motors at a low power setting which is high enough for them to rotate easily, such as 30%
Note that the fault state may be true at power up, this is normal and should clear when both motors have been driven
If there are no faults but you cannot make your motors move check GetEpo to see if the safety switch has been tripped
For more details check the website at www.piborg.org/picoborgrev and double check the wiring instructions
        """ 
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_DRIVE_FAULT, I2C_MAX_LEN)
            print "GetDriveFault::",i2cRecv
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading the drive fault state!')
            return

        if i2cRecv[1] == COMMAND_VALUE_OFF:
            return False
        else:
            return True

    def MotorsOff(self):
        """
MotorsOff()

Sets all motors to stopped, useful when ending a program
        """
        try:
            self.bus.write_byte_data(self.i2cAddress, COMMAND_ALL_OFF, 0)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motors off command!')

    def SetEncoderMoveMode(self, state):
        """
SetEncoderMoveMode(state)

Sets the system to enable or disable the encoder based move mode
In encoder move mode (enabled) the EncoderMoveMotor* commands are available to move fixed distances
In non-encoder move mode (disabled) the SetMotor* commands should be used to set drive levels
The encoder move mode requires that the encoder feedback is attached to an encoder signal, see the website at www.piborg.org/picoborgrev for wiring instructions
The encoder based move mode is disabled at power on
        """
        if state:
            level = COMMAND_VALUE_ON
        else:
            level = COMMAND_VALUE_OFF

        try:
            self.bus.write_byte_data(self.i2cAddress, COMMAND_SET_ENC_MODE, level)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending the encoder move mode!')


    def SetSpeedMotor(self, power):
        """
SetSpeedMotor(power)

 velocity pulse/seg
        """
        command=COMMAND_SET_SPEED_FBCK
        pwr=power
        if pwr < 0:
            pwr= 65535+pwr
            
        velLow = pwr & 0xFF
        velHigh = (pwr >> 8) & 0xFF

        #print 'SetSpeedMotor=',pwr
        try:
            self.bus.write_i2c_block_data(self.i2cAddress, command, [velHigh, velLow])
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending motor 1 drive level!')



    def GetEncoderSpeed(self):
        """
power = GetEncoderSpeed()
pulse/seg
        """
        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, COMMAND_GET_SPEED_FBCK, I2C_MAX_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading motor encoder move speed limit!')
            return

        #print 'i2cRecv[1]=',i2cRecv[1],', i2cRecv[2]=',i2cRecv[2]
#        power = float(i2cRecv[1]) / float(PWM_MAX)
        power = int(i2cRecv[1]*256)+int(i2cRecv[2])
        #print 'GetEncoderSpeed=',power
	if power > 32767:
		power= power-65535

        return power

    def SetPIDVariables(self, nameK,value):

        if nameK == 'KMOT':
            # Reverse
            command = COMMAND_SET_KMOT
        elif nameK == 'KP':
            # Reverse
            command = COMMAND_SET_KP
        elif nameK == 'KI':
            # Reverse
            command = COMMAND_SET_KI    
            
        val = int(value*10)

        try:
            self.bus.write_byte_data(self.i2cAddress, command, val)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed sending PID Variable!')

    def GetPIDVariables(self, nameK):

        if nameK == 'KMOT':
            # Reverse
            command = COMMAND_GET_KMOT
        elif nameK == 'KP':
            # Reverse
            command = COMMAND_GET_KP
        elif nameK == 'KI':
            # Reverse
            command = COMMAND_GET_KI    

        try:
            i2cRecv = self.bus.read_i2c_block_data(self.i2cAddress, command, I2C_MAX_LEN)
        except KeyboardInterrupt:
            raise
        except:
            self.Print('Failed reading PID Variable!')
            return

        value = float(i2cRecv[1]) 

        value=float(value/10.0)
        return value



    def Help(self):
        """
Help()

Displays the names and descriptions of the various functions and settings provided
        """
        funcList = [PicoBorgRev.__dict__.get(a) for a in dir(PicoBorgRev) if isinstance(PicoBorgRev.__dict__.get(a), types.FunctionType)]
        funcListSorted = sorted(funcList, key = lambda x: x.func_code.co_firstlineno)

        print self.__doc__
        print
        for func in funcListSorted:
            print '=== %s === %s' % (func.func_name, func.func_doc)

